/*
 * -------------------------------------------------------------------------
 * i2c-ibm_iic.c Support for the IIC peripheral on PPC 4xx
 * -------------------------------------------------------------------------
 *
 * Cleaned and unified by Matt Porter <mporter@mvista.com>
 *
 * Original driver by Ian DaSilva <idasilva@mvista.com> and Armin Kuster
 * <akuster@mvista.com>
 *
 * Copyright 2000-2003 MontaVista Software Inc.
 *
 * ----------------------------------------------------------------------------
 * This file was highly leveraged from i2c-elektor.c, which was created
 * by Simon G. Vogl and Hans Berglund:
 *
 *
 * Copyright 1995-97 Simon G. Vogl
 *           1998-99 Hans Berglund
 *
 * With some changes from Kysti Mlkki <kmalkki@cc.hut.fi> and even
 * Frodo Looijaard <frodol@dds.nl>
 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * ----------------------------------------------------------------------------
 */

#include <linux/kernel.h>
#include <linux/ioport.h>
#include <linux/module.h>
#include <linux/delay.h>
#include <linux/slab.h>
#include <linux/version.h>
#include <linux/init.h>
#include <asm/irq.h>
#include <asm/io.h>
#include <linux/i2c.h>
#include <linux/i2c-id.h>
#include <asm/ocp.h>

/*
 * This next section is configurable, and it is used to set the number
 * of i2c controllers in the system.  The default number of instances is 1,
 * however, this should be changed to reflect your system's configuration.
 */

static int i2c_debug = 0;
static int iic_scan = 0;	/* have a look at what's hanging 'round */
static wait_queue_head_t iic_wait[2];
static int iic_pending;

#define DEB(x) if (i2c_debug>=1) x
#define DEB2(x) if (i2c_debug>=2) x
#define DEB3(x) if (i2c_debug>=3) x	/* print several statistical values */
#define DEBPROTO(x) if (i2c_debug>=9) x;
/* debug the protocol by showing transferred bits */
#define DEF_TIMEOUT 5

#define IIC_DEV(x)	((struct ocp_dev *)adap->data)->x

#define iic_outb( reg, val) writeb(val,&(reg))
#define iic_inb( reg) readb(&(reg))

#define IIC_SINGLE_XFER		0
#define IIC_COMBINED_XFER	1

#define IIC_ERR_LOST_ARB        -2
#define IIC_ERR_INCOMPLETE_XFR  -3
#define IIC_ERR_NACK            -1
#define IIC_TIMEOUT		100
#define IIC_RETRY		3

#define IIC_DEV(x)	((struct ocp_dev *)adap->data)->x

/*
 * Determine if there was an error in the requested transaction.
 * If so, determine the cause.
 */
int
iic_analyze_status(struct i2c_adapter *adap, int *error_code)
{
	int ret;
	struct iic_regs *iic;

	iic = (struct iic_regs *) IIC_DEV(vaddr);

	ret = iic_inb(iic->sts);
	if (ret & 0x04) {
		/* Error occurred */
		ret = iic_inb(iic->extsts);
		if (ret & 0x04) {
			/* Lost arbitration */
			*error_code = IIC_ERR_LOST_ARB;
		}
		if (ret & 0x02) {
			/* Incomplete transfer */
			*error_code = IIC_ERR_INCOMPLETE_XFR;
		}
		if (ret & 0x01) {
			/* Master transfer aborted by a NACK during the transfer of the address byte */
			*error_code = IIC_ERR_NACK;
		}
		return -1;
	}
	return 0;
}

/*
 * Generate I2C compliant 7-bit address.
 *
 * Note: 10 bit addresses are not supported in this driver, although they are
 * supported by the hardware.  This functionality needs to be implemented.
 */
static inline int
iic_do_address(struct i2c_adapter *adap, struct i2c_msg *msg, int retries)
{
	struct iic_regs *iic;
	unsigned short flags = msg->flags;
	unsigned char addr;

	iic = (struct iic_regs *) IIC_DEV(vaddr);

/*
 * The following segment for 10 bit addresses needs to be ported
 */
#if 0
	/* Ten bit addresses not supported right now */
	if ((flags & I2C_M_TEN)) {
		/* a ten bit address */
		addr = 0xf0 | ((msg->addr >> 7) & 0x03);
		DEB2(printk(KERN_DEBUG "addr0: %d\n", addr));
		/* try extended address code... */
		ret = try_address(adap, addr, retries);
		if (ret != 1) {
			printk(KERN_ERR
			       "iic_do_address: died at extended address code.\n");
			return -EREMOTEIO;
		}
		/* the remaining 8 bit address */
		iic_outb(msg->addr & 0x7f);
		/* Status check comes here */
		if (ret != 1) {
			printk(KERN_ERR
			       "iic_do_address: died at 2nd address code.\n");
			return -EREMOTEIO;
		}
		if (flags & I2C_M_RD) {
			i2c_repstart(adap);
			/* okay, now switch into reading mode */
			addr |= 0x01;
			ret = try_address(adap, addr, retries);
			if (ret != 1) {
				printk(KERN_ERR
				       "iic_do_address: died at extended address code.\n");
				return -EREMOTEIO;
			}
		}
	} else
		/* ---------->  normal 7 bit address */
		/*Ten bit addresses not supported yet */
#endif
		addr = (msg->addr << 1);
	if (flags & I2C_M_RD)
		addr |= 1;
	if (flags & I2C_M_REV_DIR_ADDR)
		addr ^= 1;
	/*
	 * Write to the low slave address
	 */
	iic_outb(iic->lmadr, addr);
	/*
	 * Write zero to the high slave register since we are
	 * only using 7 bit addresses
	 */
	iic_outb(iic->hmadr, 0);

	return 0;
}

/*
 * Initialize IBM IIC.
 */
static int
iic_adap_init(struct i2c_adapter *adap)
{
	struct iic_regs *iic;
	unsigned short retval;

	iic = (struct iic_regs *) ((struct ocp_dev *) adap->data)->vaddr;

	/* Clear master low master address */
	iic_outb(iic->lmadr, 0);

	/* Clear high master address */
	iic_outb(iic->hmadr, 0);

	/* Clear low slave address */
	iic_outb(iic->lsadr, 0);

	/* Clear high slave address */
	iic_outb(iic->hsadr, 0);

	/* Clear status */
	iic_outb(iic->sts, 0x0a);

	/* Clear extended status */
	iic_outb(iic->extsts, 0x8f);

	/* Set clock division */
	iic_outb(iic->clkdiv, 0x04);

	retval = iic_inb(iic->clkdiv);
	DEB(printk("iic_adap_init: CLKDIV register = %x\n", retval));

	/* Enable interrupts on Requested Master Transfer Complete */
	iic_outb(iic->intmsk, 0x01);

	/* Clear transfer count */
	iic_outb(iic->xfrcnt, 0x0);

	/* Clear extended control and status */
	iic_outb(iic->xtcntlss, 0xf0);

	/* Set mode control (flush master data buf, enable hold SCL, exit */
	/* unknown state.                                                 */
	iic_outb(iic->mdcntl, 0x47);

	/* Clear control register */
	iic_outb(iic->cntl, 0x0);

	DEB2(printk(KERN_DEBUG "iic_adap_init: IBM IIC initialized.\n"));
	return 0;
}

/*
 * Description: After we issue a transaction on the IIC bus, this function
 * is called.  It puts this process to sleep until we get an interrupt from
 * from the controller telling us that the transaction we requested in complete.
 */
static int
iic_wait_for_irq(struct i2c_adapter *adap, int *status)
{

	int timeout = DEF_TIMEOUT;
	int retval;
	struct iic_regs *iic;

	iic = (struct iic_regs *) IIC_DEV(vaddr);

	*status = iic_inb(iic->sts);

	while (timeout-- && (*status & 0x01)) {
		int timeout = 2;
		struct ocp_dev *iic_dev;

		iic_dev = (struct ocp_dev *) adap->data;

		/*
		 * If interrupts are enabled (which they are),
		 * then put the process to sleep.  This process
		 * will be awakened by two events -- either the
		 * the IIC peripheral interrupts or the timeout
		 * expires. 
		 */
		if (iic_dev->irq > 0) {
			cli();
			if (iic_pending == 0) {
				interruptible_sleep_on_timeout(&
							       (iic_wait
								[iic_dev->num]),
							       timeout * HZ);
			} else
				iic_pending = 0;
			sti();
		} else {
			/*
			 * If interrupts are not enabled then
			 * delay for a reasonable amount of
			 * time and return.  We expect that by
			 * time we return to the calling function
			 * that the IIC has finished our requested
			 * transaction and the status bit reflects
			 * this.
			 * 
			 * udelay is probably not the best choice
			 * for this since it is the equivalent of
			 * a busy wait.
			 */
			udelay(100);
		}
		*status = iic_inb(iic->sts);
	}

	if (timeout <= 0) {
		/* Issue stop signal on the bus, and force an interrupt */
		retval = iic_inb(iic->cntl);
		iic_outb(iic->cntl, retval | 0x80);
		/* Clear status register */
		iic_outb(iic->sts, 0x0a);
		/* Exit unknown bus state */
		retval = iic_inb(iic->mdcntl);
		iic_outb(iic->mdcntl, (retval | 0x02));

		/* Pending transfer bit cleared? */
		retval = iic_inb(iic->sts);
		retval = retval & 0x01;
		if (retval) {
			/*
			 * IIC has locked up. Perform a soft reset
			 * to regain control of the IIC and then
			 * reinitialize.
			 */
			iic_outb(iic->xtcntlss, 0x01);
			udelay(500);
			iic_adap_init(adap);
			/* Pending transfer bit cleared? */
			retval = iic_inb(iic->sts);
			retval = retval & 0x01;
			if (retval) {
				printk(KERN_CRIT
				       "IIC has locked up.  Processor reset is required.\n");
			}
			/*
			 * IIC sometimes needs the interrupt bit
			 * written again.  Sometimes it does not
			 * stick when written in iic_adap_init.
			 */
			iic_outb(iic->intmsk, 0x01);
		}
		return (-1);
	} else
		return (0);
}

/*
 * Low level master send transaction.
 */

static int
iic_sendbytes(struct i2c_adapter *adap, const char *buf,
	      int count, int xfer_flag)
{
	struct iic_regs *iic;
	int wrcount, status, timeout;
	int loops, remainder, i, j;
	int ret, error_code;

	iic = (struct iic_regs *) IIC_DEV(vaddr);

	if (count == 0)
		return 0;
	wrcount = 0;
	loops = count / 4;
	remainder = count % 4;

	if ((loops > 1) && (remainder == 0)) {
		for (i = 0; i < (loops - 1); i++) {

			/* Write four bytes to master data buffer */
			for (j = 0; j < 4; j++) {
				iic_outb(iic->mdbuf, buf[wrcount++]);
			}

			/* Issue IIC command to begin transmission */
			iic_outb(iic->cntl, 0x35);

			/*
			 * Wait for transmission to complete.  When
			 * complete, continue next loop iteration.
			 */
			timeout = iic_wait_for_irq(adap, &status);

			if (timeout < 0) {
				/* Error handling */
				return wrcount;
			}

			ret = iic_analyze_status(adap, &error_code);

			if (ret < 0) {
				if (error_code == IIC_ERR_INCOMPLETE_XFR) {
					/* Return bytes transferred */
					ret = iic_inb(iic->xfrcnt);
					ret = ret & 0x07;
					return (wrcount - 4 + ret);
				} else
					return error_code;
			}
		}
	} else if ((loops >= 1) && (remainder > 0)) {
		for (i = 0; i < loops; i++) {
			/* Write four bytes to master data buffer */
			for (j = 0; j < 4; j++) {
				iic_outb(iic->mdbuf, buf[wrcount++]);
			}

			/* Issue IIC command to begin transmission */
			iic_outb(iic->cntl, 0x35);

			/*
			 * Wait for transmission to complete.  When
			 * complete, continue next loop iteration.
			 */
			timeout = iic_wait_for_irq(adap, &status);
			if (timeout < 0)
				return wrcount;

			ret = iic_analyze_status(adap, &error_code);
			if (ret < 0) {
				if (error_code == IIC_ERR_INCOMPLETE_XFR) {
					/* Return bytes transferred */
					ret = iic_inb(iic->xfrcnt);
					ret = ret & 0x07;
					return (wrcount - 4 + ret);
				} else
					return error_code;
			}
		}
	}

	if (remainder == 0)
		remainder = 4;

	/* Write the remaining bytes (<=4) */
	for (i = 0; i < remainder; i++)
		iic_outb(iic->mdbuf, buf[wrcount++]);

	if (xfer_flag == IIC_COMBINED_XFER) {
		iic_outb(iic->cntl, (0x09 | ((remainder - 1) << 4)));
	} else {
		iic_outb(iic->cntl, (0x01 | ((remainder - 1) << 4)));
	}
	DEB2(printk(KERN_DEBUG "iic_sendbytes: Waiting for interrupt\n"));
	timeout = iic_wait_for_irq(adap, &status);
	if (timeout < 0)
		return wrcount;

	ret = iic_analyze_status(adap, &error_code);

	if (ret < 0) {
		if (error_code == IIC_ERR_INCOMPLETE_XFR) {
			/* Return bytes transferred */
			ret = iic_inb(iic->xfrcnt);
			ret = ret & 0x07;
			return (wrcount - 4 + ret);
		} else
			return error_code;
	}

	DEB2(printk(KERN_DEBUG "iic_sendbytes: Got interrupt\n"));
	return wrcount;
}

/*
 * Low level master read transaction.
 */
static int
iic_readbytes(struct i2c_adapter *adap, char *buf, int count, int xfer_type)
{
	struct iic_regs *iic;
	int rdcount = 0, i, status, timeout;
	int loops, remainder, j;
	int ret, error_code;

	iic = (struct iic_regs *) IIC_DEV(vaddr);

	if (count == 0)
		return 0;
	loops = count / 4;
	remainder = count % 4;

	if ((loops > 1) && (remainder == 0)) {
		for (i = 0; i < (loops - 1); i++) {
			/* Issue IIC master read command (<=4 bytes) */
			iic_outb(iic->cntl, 0x37);

			/*
			 * Wait for transmission to complete.  When
			 * complete, continue next loop iteration.
			 */
			timeout = iic_wait_for_irq(adap, &status);
			if (timeout < 0)
				return rdcount;

			ret = iic_analyze_status(adap, &error_code);

			if (ret < 0) {
				if (error_code == IIC_ERR_INCOMPLETE_XFR)
					return rdcount;
				else
					return error_code;
			}

			for (j = 0; j < 4; j++) {
				/*
				 * Wait for data to shuffle to top of
				 * data buffer. This value needs to be
				 * optimized.
				 */
				udelay(1);
				buf[rdcount] = iic_inb(iic->mdbuf);
				rdcount++;
			}
		}
	}

	else if ((loops >= 1) && (remainder > 0)) {
		for (i = 0; i < loops; i++) {
			/* Issue IIC master read (<=4 bytes) */
			iic_outb(iic->cntl, 0x37);

			/*
			 * Wait for transmission to complete.  When
			 * complete, continue next loop iteration.
			 */
			timeout = iic_wait_for_irq(adap, &status);
			if (timeout < 0)
				return rdcount;

			ret = iic_analyze_status(adap, &error_code);

			if (ret < 0) {
				if (error_code == IIC_ERR_INCOMPLETE_XFR)
					return rdcount;
				else
					return error_code;
			}

			for (j = 0; j < 4; j++) {
				/*
				 * Wait for data to shuffle to top of
				 * data buffer. This value needs to be
				 * optimized.
				 */
				udelay(1);
				buf[rdcount] = iic_inb(iic->mdbuf);
				rdcount++;
			}
		}
	}

	if (remainder == 0)
		remainder = 4;
	DEB2(printk
	     (KERN_DEBUG "iic_readbytes: writing %x to IIC0CNTL\n",
	      (0x03 | ((remainder - 1) << 4))));

	if (xfer_type == IIC_COMBINED_XFER) {
		iic_outb(iic->cntl, (0x0b | ((remainder - 1) << 4)));
	} else {
		iic_outb(iic->cntl, (0x03 | ((remainder - 1) << 4)));
	}
	DEB2(printk(KERN_DEBUG "iic_readbytes: Wait for pin\n"));
	timeout = iic_wait_for_irq(adap, &status);
	DEB2(printk(KERN_DEBUG "iic_readbytes: Got the interrupt\n"));
	if (timeout < 0)
		return rdcount;

	ret = iic_analyze_status(adap, &error_code);

	if (ret < 0) {
		if (error_code == IIC_ERR_INCOMPLETE_XFR)
			return rdcount;
		else
			return error_code;
	}

	for (i = 0; i < remainder; i++) {
		buf[rdcount] = iic_inb(iic->mdbuf);
		rdcount++;
	}

	return rdcount;
}

/*
 * Execute combined transations. Combined transactions consist
 * of combinations of reading and writing blocks of data. Each
 * transfer (i.e. a read or a write) is separated by a repeated
 * start condition.
 */
static int
iic_combined_transaction(struct i2c_adapter *i2c_adap, struct i2c_msg msgs[],
			 int num)
{
	int i;

	DEB2(printk(KERN_DEBUG "Beginning combined transaction\n"));
	for (i = 0; i < num; i++) {
		int ret;
		struct i2c_msg *pmsg = &msgs[i];

		if (pmsg->flags & I2C_M_RD) {

			/* Last segment needs to be terminated with a stop */
			if (i < num - 1) {
				DEB2(printk(KERN_DEBUG "This one is a read\n"));
			} else {
				DEB2(printk
				     (KERN_DEBUG "Doing the last read\n"));
			}
			ret =
			    iic_readbytes(i2c_adap, pmsg->buf, pmsg->len,
					  (i <
					   num -
					   1) ? IIC_COMBINED_XFER :
					  IIC_SINGLE_XFER);

			if (ret != pmsg->len) {
				DEB2(printk("iic_combined_transation: fail: "
					    "only read %d bytes.\n", ret));
				return i;
			} else {
				DEB2(printk
				     ("iic_combined_transaction: read %d bytes.\n",
				      ret));
			}
		} else if (!(pmsg->flags & I2C_M_RD)) {

			/* Last segment needs to be terminated with a stop */
			if (i < num - 1) {
				DEB2(printk
				     (KERN_DEBUG "This one is a write\n"));
			} else {
				DEB2(printk
				     (KERN_DEBUG "Doing the last write\n"));
			}
			ret =
			    iic_sendbytes(i2c_adap, pmsg->buf, pmsg->len,
					  (i <
					   num -
					   1) ? IIC_COMBINED_XFER :
					  IIC_SINGLE_XFER);

			if (ret != pmsg->len) {
				DEB2(printk("iic_combined_transaction: fail: "
					    "only wrote %d bytes.\n", ret));
				return i;
			} else {
				DEB2(printk
				     ("iic_combined_transaction: wrote %d bytes.\n",
				      ret));
			}
		}
	}

	/* We've successfully read as many as was requested. */
	return num;
}

/*
 * Prepare controller for a transaction and call iic_sendbytes or
 * iic_readbytes to do the work.
 */
static int
iic_xfer(struct i2c_adapter *adap, struct i2c_msg msgs[], int num)
{
	struct iic_regs *iic;
	struct i2c_msg *pmsg;
	int i = 0;
	int ret;

	iic = (struct iic_regs *) IIC_DEV(vaddr);

	pmsg = &msgs[i];

	/*
	 * Clear status register
	 */
	DEB2(printk
	     (KERN_DEBUG "iic_xfer: iic_xfer: Clearing status register\n"));
	iic_outb(iic->sts, 0x0a);

	/* Wait for any pending transfers to complete */
	DEB2(printk
	     (KERN_DEBUG
	      "iic_xfer: Waiting for any pending transfers to complete\n"));
	while ((ret = iic_inb(iic->sts)) == 0x01) {
		;
	}

	/* Flush master data buffer */
	DEB2(printk(KERN_DEBUG "iic_xfer: Clearing master data buffer\n"));
	ret = iic_inb(iic->mdcntl);
	iic_outb(iic->mdcntl, ret | 0x40);

	/* Load slave address */
	DEB2(printk(KERN_DEBUG "iic_xfer: Loading slave address\n"));
	ret = iic_do_address(adap, pmsg, adap->retries);

	/* Check to see if the bus is busy */
	ret = iic_inb(iic->extsts);

	/* Mask off the irrelevent bits */
	ret = ret & 0x70;

	/* Bus is free when BCS bits in the EXTSTS register are 0b100 */
	if (ret != 0x40)
		return IIC_ERR_LOST_ARB;

	/* Combined transaction (read/write) */
	if (num > 1) {
		DEB2(printk
		     (KERN_DEBUG "iic_xfer: Call combined transaction\n"));
		ret = iic_combined_transaction(adap, msgs, num);
	}
	/* Read only */
	else if ((num == 1) && (pmsg->flags & I2C_M_RD)) {
		/* Tell device to begin reading data from the master data */
		DEB2(printk(KERN_DEBUG "iic_xfer: Call adapter's read\n"));
		ret =
		    iic_readbytes(adap, pmsg->buf, pmsg->len, IIC_SINGLE_XFER);
	}
	/* Write only */
	else if ((num == 1) && (!(pmsg->flags & I2C_M_RD))) {
		/*
		 * Write data to master data buffers and tell our device
		 * to begin transmitting
		 */
		DEB2(printk(KERN_DEBUG "iic_xfer: Call adapter's write\n"));
		ret =
		    iic_sendbytes(adap, pmsg->buf, pmsg->len, IIC_SINGLE_XFER);
	}

	return ret;
}

static u32
iic_func(struct i2c_adapter *adap)
{
	return I2C_FUNC_I2C | I2C_FUNC_SMBUS_EMUL;
}

static struct i2c_algorithm iic_algo = {
	"IBM IIC algorithm",
	I2C_ALGO_OCP,
	iic_xfer,
	NULL,
	NULL,			/* slave_xmit           */
	NULL,			/* slave_recv           */
	NULL,			/* ioctl                */
	iic_func,		/* functionality        */
};

/*
 * Register bus structure
 */
int
iic_add_bus(struct i2c_adapter *adap)
{

	DEB2(printk
	     (KERN_DEBUG "iic_add_bus: hw routines for %s registered.\n",
	      adap->name));
	adap->id |= iic_algo.id;
	adap->algo = &iic_algo;
	adap->timeout = IIC_TIMEOUT;
	adap->retries = IIC_RETRY;

	iic_adap_init(adap);
	i2c_add_adapter(adap);

	/*
	 * Scan the I2C bus for valid 7 bit addresses
	 * (ie things that ACK on 1byte read)
	 * if i2c_debug is off we print everything on one line.
	 * if i2c_debug is on we do a newline per print so we don't
	 * clash too much with printk's in the other functions.
	 * TODO: check for 10-bit mode and never run as a slave.
	 */
	if (iic_scan) {

		int found = 0;
		int i;
		printk(KERN_INFO " iic_add_bus: scanning bus %s. Found ",
		       adap->name);
		if (i2c_debug)
			printk("\n");
		for (i = 0; i < 0x7f; i++) {
			int ret;
			struct i2c_msg msg;
			char data[1];
			msg.addr = i;
			msg.buf = data;
			msg.len = 1;
			msg.flags = I2C_M_RD;
			if ((ret = iic_xfer(adap, &msg, 1)) == 1) {
				if (i2c_debug)
					printk
					    ("I2C Found 0x%02x ret %d Data 0x%02x\n",
					     i, ret, data[0]);
				else
					printk("0x%02x ", i);
				found++;
			}
		}
		if (i2c_debug) {
			if (!found)
				printk("I2C Found Nothing\n");
		} else {
			if (found)
				printk("\n");
			else
				printk("Nothing\n");
		}

	}
	return 1;
}

int
iic_del_bus(struct i2c_adapter *adap)
{
	int res;
	if ((res = i2c_del_adapter(adap)) < 0)
		return res;
	DEB2(printk
	     (KERN_DEBUG "iic_del_bus: adapter unregistered: %s\n",
	      adap->name));

	return 1;
}

/*
 * IIC interrupt handler
 */
static void
iic_handler(int this_irq, void *dev_id, struct pt_regs *regs)
{
	int ret;
	struct iic_regs *iic;
	struct ocp_dev *iic_dev;

	iic_dev = (struct ocp_dev *) dev_id;
	iic = (struct iic_regs *) iic_dev->vaddr;

	iic_pending = 1;

	DEB2(printk("iic_handler: in interrupt handler\n"));
	/* Read status register */
	ret = readb((int) &(iic->sts));
	DEB2(printk("iic_handler: status = %x\n", ret));
	/* Clear status register */
	writeb(0x0a, (int) &(iic->sts));
	wake_up_interruptible(&(iic_wait[iic_dev->num]));
}

/*
 * Release irq and memory
 */
static void
iic_release(void)
{
	int i;
	struct ocp_dev *iic_drv;

	for (i = 0; i < ocp_get_max(IIC); i++) {
		iic_drv = ocp_get_dev(IIC, i);
		if (iic_drv->irq > 0) {
			disable_irq(iic_drv->irq);
			free_irq(iic_drv->irq, 0);
		}
	}

}

static int
iic_reg(struct i2c_client *client)
{
	return 0;
}

static int
iic_unreg(struct i2c_client *client)
{
	return 0;
}

/*
 * Increment module usage count
 */
static void
iic_inc_use(struct i2c_adapter *adap)
{
#ifdef MODULE
	MOD_INC_USE_COUNT;
#endif
}

/*
 * Decrement module usage count
 */
static void
iic_dec_use(struct i2c_adapter *adap)
{
#ifdef MODULE
	MOD_DEC_USE_COUNT;
#endif
}

/*
 * Called when the module is loaded.  This function starts the
 * cascade of calls up through the heirarchy of i2c modules
 * (i.e. up to the algorithm layer and into to the core layer)
 */
static int __init
iic_init(void)
{
	struct ocp_dev *iic_drv;
	struct i2c_adapter *adap;
	int curr_iic = 0;

	printk(KERN_INFO "IBM IIC driver\n");

	while (curr_iic != -ENXIO) {
		if (!(iic_drv = ocp_alloc_dev(sizeof (struct i2c_adapter))))
			return -ENOMEM;
		iic_drv->type = IIC;
		if ((curr_iic = ocp_register(iic_drv)) != -ENXIO) {
			adap = (struct i2c_adapter *) iic_drv->ocpdev;
			iic_drv->vaddr = ioremap(iic_drv->paddr, 17);
			strcpy(adap->name, "IBM IIC adapter");
			adap->data = (void *) iic_drv;
			adap->id = I2C_HW_OCP;
			adap->algo = NULL;
			adap->inc_use = iic_inc_use;
			adap->dec_use = iic_dec_use;
			adap->client_register = iic_reg;
			adap->client_unregister = iic_unreg;

			init_waitqueue_head(&(iic_wait[curr_iic]));

			if (iic_drv->irq > 0) {
				if (request_irq
				    (iic_drv->irq,
				     iic_handler, 0, "IBM IIC", iic_drv)) {
					printk(KERN_ERR
					       "iic_init: Request irq %d failed\n",
					       iic_drv->irq);
					iic_drv->irq = 0;
				} else {
					DEB3(printk
					     ("iic_init: Enabled interrupt\n"));
				}
			}

			if (iic_add_bus(adap) < 0)
				return -ENODEV;

		} else {
			ocp_free_dev(iic_drv);
			break;
		}		/* end if */
	}			/* end while */
	return (curr_iic == -ENXIO) ? 0 : curr_iic;
}

static void
iic_exit(void)
{
	int i;
	struct ocp_dev *iic_drv;
	struct i2c_adapter *adap;

	for (i = 0; i < ocp_get_max(IIC); i++) {
		iic_drv = ocp_get_dev(IIC, i);
		adap = (struct i2c_adapter *) iic_drv->ocpdev;
		iic_del_bus(adap);
	}

	iic_release();
}

/*
 * Put the process to sleep for a period equal to timeout 
 */
static inline void
iic_sleep(unsigned long timeout)
{
	schedule_timeout(timeout * HZ);
}

MODULE_AUTHOR("MontaVista Software <www.mvista.com>");
MODULE_DESCRIPTION("IBM IIC driver");
MODULE_PARM(iic_scan, "i");
MODULE_PARM(i2c_debug, "i");
MODULE_PARM_DESC(iic_scan, "Scan for active chips on the bus");
MODULE_PARM_DESC(i2c_debug,
		 "debug level - 0 off; 1 normal; 2,3 more verbose; 9 iic-protocol");
MODULE_LICENSE("GPL");

module_init(iic_init);
module_exit(iic_exit);
